@@ -12,6 +12,7 @@ from b8 import ui
 
 
 class CwdDiscovery(GObject.GObject):
+  """Discover the working directory of a running process."""
 
   __gtype_name__ = 'b8-terminals-cwddiscovery'
 
@@ -36,7 +37,12 @@ class CwdDiscovery(GObject.GObject):
     self.running = False
 
   def refresh(self):
-    l = Gio.File.new_for_path(self.proclink).query_info('*', 0, None)
+    try:
+     l = Gio.File.new_for_path(self.proclink).query_info('*', 0, None)
+    except GLib.Error:
+      self.stop()
+      return False
+      
     new = Gio.File.new_for_path(l.get_symlink_target())
     if not self.cwd or (new.get_path() != self.cwd.get_path()):
       self.cwd = new
@@ -52,17 +58,23 @@ class CwdDiscovery(GObject.GObject):
 
 class TerminalTheme(GObject.GObject):
 
-  def __init__(self):
+  __gtype_name__ = 'b8-terminal-theme'
+
+  foreground = GObject.Property(type=Gdk.RGBA)
+  background = GObject.Property(type=Gdk.RGBA)
+  cursor = GObject.Property(type=Gdk.RGBA)
+  font = GObject.Property(type=Pango.FontDescription)
+
+  def __init__(self, name, font):
     GObject.GObject.__init__(self)
     self.default_theme = TERMINAL_THEMES.get('b8')
-    #theme_name = self.b8.config.get('terminal', 'theme')
-    raw_theme = TERMINAL_THEMES.get('b8')
+    raw_theme = TERMINAL_THEMES.get(name)
     if not raw_theme:
       self.error(f'theme {theme_name} does not exist, using default')
       raw_theme = self.default_theme
     self.parse(raw_theme)
-    self.font_name = 'Monospace 13'
-    self.font_desc = Pango.font_description_from_string(self.font_name)
+    self.font_name = font
+    self.font = Pango.font_description_from_string(self.font_name)
 
   def parse(self, theme):
     self.foreground = Gdk.RGBA()
@@ -78,6 +90,12 @@ class TerminalTheme(GObject.GObject):
       c.parse(s)
       self.palette.append(c)
 
+  def apply(self, t: Vte.Terminal):
+    t.set_colors(self.foreground, self.background, self.palette)
+    t.set_color_cursor(self.cursor)
+    t.set_color_cursor_foreground(self.cursor)
+    t.set_font(self.font)
+
 
 class Terminals(Gtk.Notebook, ui.MenuHandlerMixin):
 
@@ -90,21 +108,20 @@ class Terminals(Gtk.Notebook, ui.MenuHandlerMixin):
     'directory-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
   }
 
-  def __init__(self, font):
+  def __init__(self, font, theme):
     Gtk.Notebook.__init__(self)
-    self.theme = TerminalTheme()
     self.set_tab_pos(Gtk.PositionType.BOTTOM)
     self.set_scrollable(True)
-    self.font = Pango.font_description_from_string(font)
+    self.theme = TerminalTheme(theme, font)
 
   def create(self, wd=None):
     if not wd:
       wd = os.path.expanduser('~')
-    t = Terminal(self)
-    t.term.set_font(self.font)
+    t = Terminal()
+    self.theme.apply(t.term)
     t.start(wd)
     self.append(t)
-    #self.pack_start(t, True, True, 0)
+    t.term.grab_focus()
 
   def append(self, t):
     pagenum = self.append_page(t, t.label)
@@ -128,36 +145,45 @@ class Terminals(Gtk.Notebook, ui.MenuHandlerMixin):
   def change_tab(self, n):
     self.set_current_page(n)
     p = self.get_nth_page(n)
-    if hasattr(p, 'term'):
-      p = p.term
-    p.grab_focus()
-    
+    p.term.grab_focus()
+
+  def remove_terminal(self, t):
+    self.remove(t)
+    if not self.get_n_pages():
+      self.create()
+
+  def shutdown(self):
+    for child in self.get_children():
+      child.term.disconnect(child.exited_handler)
+      child.cwd_discovery.stop()
+      if child.pid > 0:
+        os.kill(child.pid, 15)
+
 
 
 class Terminal(Gtk.HBox):
 
   __gtype_name__ = 'b8-terminals-terminal'
 
-
   file_match = GObject.Property(type=int)
   url_match = GObject.Property(type=int)
   cwd_discovery = GObject.Property(type=CwdDiscovery)
   label = GObject.Property(type=Gtk.Label)
   pid = GObject.Property(type=int)
   cwd = GObject.Property(type=Gio.File)
+  exited_handler = GObject.Property(type=int)
 
-  def __init__(self, parent):
+  def __init__(self):
     Gtk.HBox.__init__(self)
-    self.parent = parent
     self.term = Vte.Terminal()
     sw = Gtk.ScrolledWindow()
     sw.add(self.term)
     self.pack_start(sw, True, True, 0)
     self.pack_start(self._create_toolbar(), False, False, 0)
     self._add_matches()
-    self._set_theme(parent.theme)
     self.label = self._create_tab_label()
     self.term.connect('button-press-event', self._on_button_press_event)
+    self.exited_handler = self.term.connect('child-exited', self._on_child_exited)
 
   def _create_tab_label(self):
     self.label = Gtk.Label()
@@ -168,12 +194,6 @@ class Terminal(Gtk.HBox):
   def _update_label(self):
     self.label.set_markup(self._markup())
 
-  def _set_theme(self, theme):
-    self.term.set_colors(theme.foreground, theme.background, theme.palette)
-    self.term.set_color_cursor(theme.cursor)
-    self.term.set_color_cursor_foreground(theme.cursor)
-    self.term.set_font(theme.font_desc)
-
   def _markup(self):
     ecwd = GLib.markup_escape_text(self.cwd.get_path())
     return f'<span size="small">{ecwd} [<span weight="bold">{self.pid}</span>]</span>'
@@ -214,7 +234,6 @@ class Terminal(Gtk.HBox):
 
 
   def _on_toolbar_clicked(self, w, b, key):
-    print(b, key)
     handlers = {
         'browse': self._on_browse_clicked,
         'terminal': self._on_terminal_clicked,
@@ -227,7 +246,7 @@ class Terminal(Gtk.HBox):
       f(b)
 
   def _on_browse_clicked(self, w):
-    self.parent.emit('directory-activated', self.cwd)
+    self.get_parent().emit('directory-activated', self.cwd)
 
   def _on_copy_clicked(self, w):
     self.term.copy_clipboard_format(Vte.Format.TEXT)
@@ -239,7 +258,7 @@ class Terminal(Gtk.HBox):
     self.term.select_all()
 
   def _on_terminal_clicked(self, w):
-    self.parent.emit('terminal-activated', self.cwd)
+    self.get_parent().emit('terminal-activated', self.cwd)
 
   def _add_matches(self):
     self.url_match = self.term.match_add_regex(URL_RE, 0)
@@ -252,10 +271,6 @@ class Terminal(Gtk.HBox):
     self.cwd_discovery = CwdDiscovery(pid)
     self.cwd_discovery.connect('cwd-changed', self._on_cwd_changed)
     self.cwd_discovery.start()
-    self.term.watch_child(pid)
-    #self._update_label()
-    #self.update_label()
-    #GLib.timeout_add(1000, self.label_updater)
     self.grab_focus()
 
   def _on_cwd_changed(self, w, cwd):
@@ -268,7 +283,7 @@ class Terminal(Gtk.HBox):
         wd,
         [self._get_default_shell()],
         [],
-        GLib.SpawnFlags.DEFAULT | GLib.SpawnFlags.DO_NOT_REAP_CHILD,
+        GLib.SpawnFlags.DO_NOT_REAP_CHILD,
         None,
         None,
         -1,
@@ -316,22 +331,36 @@ class Terminal(Gtk.HBox):
       return
     if f.query_file_type(0, None) == Gio.FileType.DIRECTORY:
       if event.button == Gdk.BUTTON_PRIMARY:
-        self.parent.emit('directory-activated', f)
+        self.get_parent().emit('directory-activated', f)
       elif event.button == Gdk.BUTTON_SECONDARY:
         menu = ui.DirectoryPopupMenu(f)
-        menu.connect('activate', self.parent._on_menu_activate)
+        menu.connect('activate', self.get_parent()._on_menu_activate)
         menu.popup(event)
     else:
       if event.button == Gdk.BUTTON_PRIMARY:
-        self.parent.emit('file-activated', f)
+        self.get_parent().emit('file-activated', f)
       elif event.button == Gdk.BUTTON_SECONDARY:
         menu = ui.FilePopupMenu(f)
-        menu.connect('activate', self.parent._on_menu_activate)
+        menu.connect('activate', self.get_parent()._on_menu_activate)
         menu.popup(event)
 
   def _on_match_url_click(self, m, event):
     webbrowser.open(m)
 
+  def _on_child_exited(self, t, status):
+    self.pid = -1
+    t.feed(f'\x1b[0;1;34mExited, '
+                   f'status: \x1b[0;1;31m{status} \r\n'
+                   f'\x1b[0mpress enter to close'.encode('utf-8'))
+    t.connect('key-press-event', self._on_keypress_after_exit)
+
+  def _on_keypress_after_exit(self, t, event):
+    key_name = Gdk.keyval_name(event.keyval)
+    if key_name == 'Return':
+      self.close()
+
+  def close(self):
+    self.get_parent().remove_terminal(self)
 
 
 # https://github.com/luvit/pcre2/blob/master/src/pcre2.h.in
